# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from biosteam.process_tools import UnitGroup
from biorefineries import corn as cn
from biorefineries.sugarcane import ConventionalEthanolTEA

__all__ = ('create_tea', 'tea_summary')

def create_tea(system, cls=ConventionalEthanolTEA):
    return cls(system, IRR=0.15,
               duration=(2018, 2038),
               depreciation='MACRS10', income_tax=0.35,
               operating_days=330, lang_factor=3,
               construction_schedule=(0.4, 0.6), WC_over_FCI=0.05,
               labor_cost=1e6, fringe_benefits=0.4,
               property_tax=0.001, property_insurance=0.005,
               supplies=0.20, maintenance=0.01, administration=0.005)

def tea_summary():
    operating_days = cn.corn_tea.operating_days
    f = operating_days * 24
    ug = UnitGroup('biorefinery', cn.corn_tea.units)
    power_utilities = ug.power_utilities
    heat_utilities = ug.heat_utilities
    heating_utilities = [i for i in heat_utilities if i.duty > 0]
    cooling_utilities = [i for i in heat_utilities if i.duty < 0]
    dct = {
        'Material costs':
            {'Corn': f * cn.corn.cost ,
             'Denaturant': f * cn.denaturant.cost,
             'Enzymes': f * (cn.alpha_amylase.cost + cn.gluco_amylase.cost),
             'Yeast': f * cn.yeast.cost,
             'Other': f * sum([i.cost for i in (cn.ammonia, cn.sulfuric_acid, cn.lime)])},
         'Utilities':
             {'Electricity':f * sum([i.cost for i in power_utilities]),
              'Steam': f * sum([i.cost for i in heating_utilities]),
              'Natural gas': f * cn.D610.natural_gas_cost},
              'Cooling water': f * sum([i.cost for i in cooling_utilities]),
         'Depreciation': cn.corn_tea.TDC / 10.,
         'Fixed operating costs': cn.corn_tea.VOC,
    }
    dct['Total production cost'] = sum([*dct['Material costs'].values(),
                                        *dct['Utilities'].values(),
                                        dct['Depreciation'],
                                        dct['Fixed operating costs']])
    return dct


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import matplotlib.pyplot as plt
import biosteam as bst
from biorefineries import lipidcane as lc
from biosteam.evaluation import Model, Metric
from biosteam.evaluation.evaluation_tools import triang
from biosteam.process_tools import UnitGroup
from biosteam import plots
import pandas as pd
import numpy as np

np.random.seed(1234)

# %% Create metrics 

ethanol_density = 2.9866 # kg/gallon
tea = lc.lipidcane_tea
ugroup = UnitGroup('Biorefinery', tea.units)
get_steam_demand = lambda: sum([i.flow for i in lc.BT.steam_utilities]) * 18.01528 * tea._operating_hours / 1000
get_electricity_consumption = ugroup.get_electricity_consumption
get_electricity_production = ugroup.get_electricity_production
get_excess_electricity = lambda: get_electricity_production() - get_electricity_consumption()
get_ethanol_production = lambda: lc.ethanol.F_mass * tea._operating_hours / 907.185
get_biodiesel_production = lambda: lc.biodiesel.F_mass * tea._operating_hours / 907.185
get_MESP = lambda: tea.solve_price(lc.ethanol) * ethanol_density
get_MFPP = lambda: tea.solve_price(lc.lipidcane)
get_IRR = lambda: tea.solve_IRR() * 100
get_FCI = lambda: tea.FCI / 10**6

metrics = (Metric('IRR', get_IRR, '%'),
           Metric('MFPP', get_MFPP, 'USD/kg'),
           Metric('MESP', get_MESP, 'USD/gal'), 
           Metric('FCI', get_FCI, 'million USD'),
           Metric('Biodiesel production', get_biodiesel_production, 'ton/yr'),
           Metric('Ethanol production', get_ethanol_production, 'ton/yr'),
           Metric('Steam demand', get_steam_demand, 'MT/yr'),
           Metric('Consumed electricity', get_electricity_consumption, 'MW'),
           Metric('Excess electricity', get_excess_electricity, 'MW'))

# Used to index metrics by name
metrics_by_name = {i.name: i for i in metrics}

# %% Create model and populate parameters

#: [float] Plant size process specification in ton / yr
plant_size_ = lc.lipidcane.F_mass * lc.lipidcane_tea.operating_days * 24 / 907.185
constant_lipid_content = False
def load_specifications():
    # Plant size
    lc.lipidcane.F_mass = plant_size_ / (lc.lipidcane_tea.operating_days * 24 / 907.185)

model = Model(lc.lipidcane_sys, metrics, load_specifications)
param = model.parameter
lipid_content = 100. * lc.utils.get_lipid_fraction()

@param(units='dry wt. %', 
       distribution=triang(lipid_content),
       baseline=lipid_content)
def set_lipic_content(lipid_content):
    if constant_lipid_content: return
    lc.utils.set_lipid_fraction(lipid_content / 100.)

@param(units='ton/yr',
       distribution=triang(plant_size_),
       baseline=plant_size_)
def set_plant_size(plant_size):
    global plant_size_
    plant_size_ = plant_size

@param(units='days/yr', 
       distribution=triang(tea.operating_days),
       baseline=tea.operating_days)
def set_operating_days(operating_days):
    tea.operating_days = operating_days
    
@param(element=lc.ethanol, units='USD/kg',
       distribution=triang(lc.ethanol.price),
       baseline=lc.ethanol.price)
def set_ethanol_price(price):
    lc.ethanol.price = price
    
@param(element=lc.lipidcane, units='USD/kg',
       distribution=triang(lc.lipidcane.price),
       baseline=lc.lipidcane.price)
def set_lipidcane_price(price):
    lc.lipidcane.price = price

@param(element=bst.PowerUtility, units='USD/kWhr',
       distribution=triang(bst.PowerUtility.price),
       baseline=bst.PowerUtility.price)
def set_electricity_price(price):
    bst.PowerUtility.price = price

@param(units='%',
       distribution=triang(100. * tea.IRR),
       baseline=100. * tea.IRR)
def set_IRR(IRR):
    tea.IRR = IRR / 100.

model.get_parameters()[-1].name = 'IRR'

# %% Interface

def evaluate_sample(lipid_content, plant_size, operating_days, ethanol_price,
                    lipidcane_price, electricity_price, IRR):
    """
    Return a pandas.Series object with metric values.
    
    Examples
    --------
    >>> # baseline_sample = model.get_baseline_sample()
    >>> # baseline_sample
    array([1.012e+01, 1.764e+06, 2.000e+02, 7.890e-01, 3.455e-02, 7.820e-02, 1.804e+01])
    >>> # metric_values = evaluate_sample(*baseline_sample)
    >>> # metric_values
    IRR [%]                                     0.18
    MFPP [USD/kg]                               0.0346
    MESP [USD/gal]                              2.36
    FCI [million USD]                           222
    Biodiesel production [ton/yr]               4.62e+04
    Ethanol production [ton/yr]                 6.14e+04
    Steam demand [MT/yr]                        5.67e+05
    Consumed electricity [MW]                   4.63e+04
    Excess electricity [MW]                     3.98e+05
    dtype: float64
    >>> metric_values['Consumed electricity [MW]']
    46295.52888978263
    
    """
    return model([lipid_content, plant_size, operating_days, ethanol_price,
                  lipidcane_price, electricity_price, IRR])['Biorefinery']

def plot_montecarlo(metric, N=100):
    """
    Create a matplotlib boxplot and return plot objects.
    
    Examples
    --------
    >>> # plot_montecarlo('Excess electricity', 100) -> creates plot
    {'whiskers': [<matplotlib.lines.Line2D at 0x27ac8469b88>,
      <matplotlib.lines.Line2D at 0x27ab70a9c08>],
     'caps': [<matplotlib.lines.Line2D at 0x27ac852ed88>,
      <matplotlib.lines.Line2D at 0x27ac8551f48>],
     'boxes': [<matplotlib.patches.PathPatch at 0x27ac84692c8>],
     'medians': [<matplotlib.lines.Line2D at 0x27ac855b2c8>],
     'fliers': [<matplotlib.lines.Line2D at 0x27ac8551fc8>],
     'means': []}
    
    """
    if isinstance(metric, str): metric = metrics_by_name[metric]
    samples = model.sample(N, 'L')
    model.load_samples(samples)
    model.evaluate()
    df = model.table[[metric.index]]
    bx = plots.plot_montecarlo(df, transpose=True)
    plt.xticks([], [])
    plt.ylabel(metric.name_with_units)
    return bx

def plot_montecarlo_across_lipid_content(metric,
                                         N_coordinate_samples = 50,
                                         N_coordinates = 10,
                                         rule = 'L'):
    """
    Create a montecarlo plot across lipid content and return figure.
    
    Examples
    --------
    >>> # plot_montecarlo_across_lipid_content('IRR') -> creates plot
    
    """
    global constant_lipid_content
    if isinstance(metric, str): metric = metrics_by_name[metric]
    
    ### Evaluate ###
    
    # Monte Carlo across lipid fraction    
    coordinate = np.linspace(0.11, 0.01, N_coordinates)
    samples = model.sample(N_coordinate_samples, rule)
    model.load_samples(samples)
    
    constant_lipid_content = True
    model.evaluate_across_coordinate('Lipid fraction',
          lc.utils.set_lipid_fraction, coordinate,
          xlfile='Monte Carlo across lipid fraction.xlsx')
    constant_lipid_content = False    
    
    # TODO: Add model for sugarcane
    # # Sugar cane Monte Carlo    
    # samples = model_sc.sample(N_coordinate_samples, rule)
    # model_sc.load_samples(samples)
    # model_sc.evaluate()
    # model_sc.table.to_excel('Monte Carlo sugarcane.xlsx')

    ### Plot ###

    fig = plt.figure()
    
    # Plot lipid-cane    
    readxl = lambda sheet: pd.read_excel('Monte Carlo across lipid fraction.xlsx',
                                         sheet_name=sheet, index_col=0)
    df_metric = readxl(metric.name)
    lipid_content = np.array(df_metric.columns) * 100
    plt.ylabel(metric.name_with_units)
    plots.plot_montecarlo_across_coordinate(lipid_content, df_metric)
    plt.xlim([1, 11])
    plt.xlabel("Lipid content [dry wt. %]")
    
    # # Plot sugarcane
    # data_sc = pd.read_excel('Monte Carlo sugarcane.xlsx', header=[0,1])
    # get_metric = lambda name: np.asarray(data_sc['Biorefinery', name]).flatten()
    # metric_values = get_metric(metric.name)
    # plots.plot_montecarlo(metric_values)
    # plots.plot_vertical_line(1)
    # plt.xlim(-1, 10.5)
    
    return fig



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import numpy as np
import biosteam as bst
import thermosteam as tmo
import flexsolve as flx
import pytest
from biosteam.process_tools import UnitGroup
from importlib import import_module

__all__ = (
    'test_sugarcane',
    'test_lipidcane',
    'test_cornstover',
    'test_LAOs',
    'test_wheatstraw',
    'test_animal_bedding',
    'test_lactic',
    'test_ethanol_adipic',
    'generate_code',
    'print_results',
)

# Block speed-up for consistent testing. Note that speed-up wrapps functions,
# thus preventing code coverage to be analyzed
bst.speed_up = tmo.speed_up = flx.speed_up = lambda: None 

feedstocks_by_module = {
    'lipidcane': 'lipidcane',
    'cornstover': 'cornstover',
    'sugarcane': 'sugarcane',
    'LAOs': 'glucose',
    'wheatstraw': 'wheatstraw',
    'animal_bedding': 'bedding',
    'lactic': 'feedstock',
    'ethanol_adipic': 'feedstock',
}
products_by_module = {
    'lipidcane': 'ethanol',
    'cornstover': 'ethanol',
    'sugarcane': 'ethanol',
    'LAOs': 'octene',
    'wheatstraw': 'ethanol',
    'animal_bedding': 'ethanol',
    'lactic': 'lactic_acid',
    'ethanol_adipic': 'ethanol',
}

def generate_code(module_name, feedstock_name=None, product_name=None):
    if not feedstock_name:
        feedstock_name = feedstocks_by_module[module_name]
    if not product_name:
        product_name = products_by_module[module_name]
    bst.process_tools.default_utilities()
    module = import_module('biorefineries.' + module_name)
    try: module.load()
    except: pass
    feedstock = getattr(module, feedstock_name)
    product = getattr(module, product_name)
    tea_name = f'{module_name}_tea'
    try:
        tea = getattr(module, tea_name)
    except AttributeError:
        try:
            tea_name = f'{feedstock_name}_tea'
            tea = getattr(module, tea_name)
        except AttributeError:
            tea_name = f'{product_name}_tea'
            tea = getattr(module, tea_name)
    units = UnitGroup('Biorefinery', tea.units)
    IRR = tea.IRR
    sales = tea.sales
    material_cost = tea.material_cost
    installed_equipment_cost = tea.installed_equipment_cost
    utility_cost = tea.utility_cost
    heating_duty = units.get_heating_duty()
    cooling_duty = units.get_cooling_duty()
    electricity_consumption = units.get_electricity_consumption()
    electricity_production = units.get_electricity_production()
    print(f"""def test_{module_name}():
    bst.process_tools.default_utilities()
    from biorefineries import {module_name} as module
    try: module.load()
    except: pass
    feedstock = module.{feedstock_name}
    product = module.{product_name}
    tea = module.{tea_name}
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, {IRR}, rtol=1e-2)
    assert np.allclose(feedstock.price, {feedstock.price}, rtol=1e-2)
    assert np.allclose(product.price, {product.price}, rtol=1e-2)
    assert np.allclose(tea.sales, {sales}, rtol=1e-2)
    assert np.allclose(tea.material_cost, {material_cost}, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, {installed_equipment_cost}, rtol=1e-2)
    assert np.allclose(tea.utility_cost, {utility_cost}, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), {heating_duty}, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), {cooling_duty}, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), {electricity_consumption}, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), {electricity_production}, rtol=1e-2)
    """
    )

def print_results(tea):
    units = UnitGroup('Biorefinery', tea.units)
    print('Sales:', tea.sales)
    print('Material cost:', tea.material_cost)
    print('Installed equipment cost:', tea.installed_equipment_cost)
    print('Utility cost:', tea.utility_cost)
    print('Heating duty:', units.get_heating_duty())
    print('Cooling duty:', units.get_cooling_duty())
    print('Electricity consumption:', units.get_electricity_consumption())
    print('Electricity production:', units.get_electricity_production())
    
def test_sugarcane():
    bst.process_tools.default_utilities()
    from biorefineries import sugarcane as module
    try: module.load()
    except: pass
    feedstock = module.sugarcane
    product = module.ethanol
    tea = module.sugarcane_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.11996491643059493, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.03455, rtol=1e-2)
    assert np.allclose(product.price, 0.789, rtol=1e-2)
    assert np.allclose(tea.sales, 87957692.44497703, rtol=1e-2)
    assert np.allclose(tea.material_cost, 60205997.48173371, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 168941419.06001467, rtol=1e-2)
    assert np.allclose(tea.utility_cost, -10153951.657158965, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 261.68940213569323, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 329.5514436234315, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 9.857678792173383, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 42.40239564204188, rtol=1e-2)

def test_lipidcane():
    bst.process_tools.default_utilities()
    from biorefineries import lipidcane as module
    try: module.load()
    except: pass
    feedstock = module.lipidcane
    product = module.ethanol
    tea = module.lipidcane_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1938986786410724, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.03455, rtol=1e-2)
    assert np.allclose(product.price, 0.789, rtol=1e-2)
    assert np.allclose(tea.sales, 102495710.65224119, rtol=1e-2)
    assert np.allclose(tea.material_cost, 61827300.90214124, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 208970766.7186907, rtol=1e-2)
    assert np.allclose(tea.utility_cost, -27425242.600098744, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 190.06061033275785, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 294.7284771066694, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 9.713095190058063, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 97.6145137801182, rtol=1e-2)

def test_cornstover():
    bst.process_tools.default_utilities()
    from biorefineries import cornstover as module
    try: module.load()
    except: pass
    feedstock = module.cornstover
    product = module.ethanol
    tea = module.cornstover_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.05158816935126135, rtol=1e-2)
    assert np.allclose(product.price, 0.7140889773406729, rtol=1e-2)
    assert np.allclose(tea.sales, 134375911.2480917, rtol=1e-2)
    assert np.allclose(tea.material_cost, 80366271.80285452, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 218887522.25337386, rtol=1e-2)
    assert np.allclose(tea.utility_cost, -8369700.152889736, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 341.617643878037, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 384.36111676640894, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 23.854539454991567, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 41.254110094061666, rtol=1e-2)
    
def test_LAOs():
    bst.process_tools.default_utilities()
    from biorefineries import LAOs as module
    try: module.load()
    except: pass
    feedstock = module.glucose
    product = module.octene
    tea = module.LAOs_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.265, rtol=1e-2)
    assert np.allclose(product.price, 1.3099409534281277, rtol=1e-2)
    assert np.allclose(tea.sales, 163427716.3748871, rtol=1e-2)
    assert np.allclose(tea.material_cost, 135661583.58974993, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 62338869.2109012, rtol=1e-2)
    assert np.allclose(tea.utility_cost, 2788634.0182542275, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 40.00054620871385, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 125.31693495458933, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 3.596591914799831, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 3.5965919147998306, rtol=1e-2)

@pytest.mark.slow
def test_wheatstraw():
    bst.process_tools.default_utilities()
    from biorefineries import wheatstraw as module
    try: module.load()
    except: pass
    feedstock = module.wheatstraw
    product = module.ethanol
    tea = module.wheatstraw_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.056999999999999995, rtol=1e-2)
    assert np.allclose(product.price, 0.9111601864648353, rtol=1e-2)
    assert np.allclose(tea.sales, 128280893.1963186, rtol=1e-2)
    assert np.allclose(tea.material_cost, 63430743.51103368, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 237017812.00864774, rtol=1e-2)
    assert np.allclose(tea.utility_cost, -5358091.605865419, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 214.26465188642763, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 261.68181864418835, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 22.95476338610469, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 34.0935726246502, rtol=1e-2)

@pytest.mark.slow
def test_animal_bedding():
    bst.process_tools.default_utilities()
    from biorefineries import animal_bedding as module
    try: module.load()
    except: pass
    feedstock = module.bedding
    product = module.ethanol
    tea = module.bedding_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.003, rtol=1e-2)
    assert np.allclose(product.price, 0.7869315431126345, rtol=1e-2)
    assert np.allclose(tea.sales, 113067734.3258351, rtol=1e-2)
    assert np.allclose(tea.material_cost, 33147675.583264522, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 279518315.8576579, rtol=1e-2)
    assert np.allclose(tea.utility_cost, -2040299.6574320346, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 157.8854022174697, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 198.23736472092472, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 32.038638068681244, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 36.280168513723815, rtol=1e-2)
    
@pytest.mark.slow
def test_lactic():
    bst.process_tools.default_utilities()
    from biorefineries import lactic as module
    try: module.load()
    except: pass
    feedstock = module.feedstock
    product = module.lactic_acid
    tea = module.lactic_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.06287583717512708, rtol=1e-2)
    assert np.allclose(product.price, 1.5290315668201484, rtol=1e-2)
    assert np.allclose(tea.sales, 332346881.2497965, rtol=1e-2)
    assert np.allclose(tea.material_cost, 219074528.09864402, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 321333077.7058257, rtol=1e-2)
    assert np.allclose(tea.utility_cost, 25050313.99229334, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 1829.09043477846, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 1895.6439958358949, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 45.390871189920496, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 0.0, rtol=1e-2)
    

@pytest.mark.slow
def test_ethanol_adipic():
    bst.process_tools.default_utilities()
    from biorefineries import ethanol_adipic as module
    try: module.load()
    except: pass
    feedstock = module.feedstock
    product = module.ethanol
    tea = module.ethanol_adipic_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.06287581915485817, rtol=1e-2)
    assert np.allclose(product.price, 0.8360250946567223, rtol=1e-2)
    assert np.allclose(tea.sales, 219891872.86398175, rtol=1e-2)
    assert np.allclose(tea.material_cost, 120581699.4153729, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 283387972.8581399, rtol=1e-2)
    assert np.allclose(tea.utility_cost, 14242922.86112985, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 296.834875396348, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 247.78343626334382, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 26.56706595101369, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 0.0, rtol=1e-2)
    
### DO NOT DELETE:
### Code commented for legacy purposes
# @pytest.mark.slow
# def test_lactic():
#     from biorefineries import lactic
#     bst.process_tools.default_utilities()
#     system = lactic.system
#     MPSP = system.lactic_acid.price
#     assert np.allclose(MPSP, 1.5285811040727408, atol=0.01)
#     tea = system.lactic_tea
#     assert np.allclose(tea.sales, 332247972.3322271, rtol=0.01)
#     assert np.allclose(tea.material_cost, 218911499.70700422, rtol=0.01)
#     assert np.allclose(tea.installed_equipment_cost, 321744540.84435904, rtol=0.01)
#     assert np.allclose(tea.utility_cost, 24995492.58140952, rtol=0.01)
#     units = UnitGroup('Biorefinery', system.lactic_tea.units)
#     assert np.allclose(system.CHP.system_heating_demand/1e6, 1763.4942302374052, rtol=0.01)
#     assert np.allclose(-system.CT.system_cooling_water_duty/1e6, 1629.0964343101657, rtol=0.01)
#     assert np.allclose(units.get_electricity_consumption(), 45.291535445041546, rtol=0.01)
#     assert np.allclose(units.get_electricity_production(), 0.0)

# @pytest.mark.slow
# def test_ethanol_adipic():
#     bst.process_tools.default_utilities()
#     from biorefineries import ethanol_adipic
#     acid = ethanol_adipic.system_acid
#     MESP = acid.ethanol.price * acid._ethanol_kg_2_gal
#     assert np.allclose(MESP, 2.486866594017598, atol=0.01)
#     tea = acid.ethanol_tea
#     assert np.allclose(tea.sales, 152002785.9542236, rtol=0.01)
#     assert np.allclose(tea.material_cost, 112973059.93909653, rtol=0.01)
#     assert np.allclose(tea.installed_equipment_cost, 202880204.7890376, rtol=0.01)
#     assert np.allclose(tea.utility_cost, -18225095.0315181, rtol=0.01)
#     units = UnitGroup('Biorefinery', acid.ethanol_tea.units)
#     assert np.allclose(acid.CHP.system_heating_demand/1e6, 333.8802418517945, rtol=0.01)
#     assert np.allclose(units.get_cooling_duty(), 327.05990128304114, rtol=0.01)
#     assert np.allclose(units.get_electricity_consumption(), 23.84999102548279, rtol=0.01)
#     assert np.allclose(units.get_electricity_production(), 55.72024685271333, rtol=0.01)
    
#     base = ethanol_adipic.system_base
#     MESP = base.ethanol.price * base._ethanol_kg_2_gal
#     assert np.allclose(MESP, 2.703699730342356, atol=0.01)
#     tea = base.ethanol_adipic_tea
#     assert np.allclose(tea.sales, 219850046.2308626, rtol=0.01)
#     assert np.allclose(tea.material_cost, 120551649.9082640, rtol=0.01)
#     assert np.allclose(tea.installed_equipment_cost, 283345546.0556234, rtol=0.01)
#     assert np.allclose(tea.utility_cost, 14241964.663629433, rtol=0.01)
#     units = UnitGroup('Biorefinery', base.ethanol_adipic_tea.units)
#     assert np.allclose(base.CHP.system_heating_demand/1e6, 296.8322623939439, rtol=0.01)
#     assert np.allclose(units.get_cooling_duty(), 247.79573940245342, rtol=0.01)
#     assert np.allclose(units.get_electricity_consumption(), 26.565278642577344, rtol=0.001)
#     assert np.allclose(units.get_electricity_production(), 0.0)
    
if __name__ == '__main__':
    test_sugarcane()
    test_lipidcane()
    test_cornstover()
    test_wheatstraw()
    test_LAOs()
    test_animal_bedding()
    test_lactic()
    test_ethanol_adipic()







# -*- coding: utf-8 -*-
"""
Created on Thu Dec 24 15:31:18 2020

@author: yrc2
"""
import flexsolve as flx
from biosteam.exceptions import InfeasibleRegion

class TiterAndInhibitorSpecification:
    __slots__ = (
        'evaporator',
        'mixer',
        'reactor',
        'product',
        'products',
        'sugars',
        'inhibitors',
        'target_titer',
        'maximum_inhibitor_concentration',
    )
    
    max_sugar_concentration = 600 # g / L
    
    def __init__(self, evaporator, mixer, reactor,
                 target_titer, product, products,
                 maximum_inhibitor_concentration=1.,
                 sugars= ('Glucose', 'Xylose'),
                 inhibitors = ('AceticAcid', 'HMF', 'Furfural')):
        self.evaporator = evaporator
        self.mixer = mixer
        self.reactor = reactor
        self.product = product
        self.products = products
        self.sugars = sugars
        self.inhibitors = inhibitors
        self.target_titer = target_titer
        self.maximum_inhibitor_concentration = maximum_inhibitor_concentration
        
    @property
    def feed(self):
        return self.evaporator.ins[0]
    
    @property
    def evaporated_sugar_solution(self):
        return self.evaporator.outs[0]
    
    @property
    def dilution_water(self):
        return self.mixer.ins[1]
    
    def run_units(self):
        self.evaporator._run()
        self.mixer._run()
        self.reactor._run()
    
    def calculate_titer(self):
        product = self.product
        return product.imass[self.products].sum() / product.F_vol
    
    def calculate_inhibitors(self): # g / L
        product = self.product
        return product.imass[self.inhibitors].sum() / product.F_vol
    
    def calculate_sugar_concentration(self): # g / L
        s = self.evaporated_sugar_solution
        return s.imass[sugars].sum() / s.F_vol 
    
    def check_sugar_concentration(self):
        if self.calculate_sugar_concentration() > self.max_sugar_concentration:
            raise InfeasibleRegion('sugar concentration')
    
    def titer_objective_function(self, V):
        self.evaporator.V = V
        self.run_units()
        return self.calculate_titer() - self.target_titer
    
    def inhibitor_objective_function(self, V):
        self.evaporator.V = V
        self.run_units()
        self.update_dilution_water()
        self.run_units()
        return self.calculate_inhibitors() - self.maximum_inhibitor_concentration
    
    def run(self):
        self.dilution_water.empty()
        self.evaporator.V = 0.
        self.run_units()
        x_titer = self.calculate_titer()
        V_min = 1e-6 
        if x_titer < self.target_titer: # Evaporate
            self.evaporator.V = V_min = flx.IQ_interpolation(self.titer_objective_function,
                                                             V_min, 1., ytol=1e-4, maxiter=100) 
        elif x_titer > self.target_titer: # Dilute
            self.update_dilution_water(x_titer)
            self.mixer._run()
            self.reactor._run()
        
        self.check_sugar_concentration()
        x_inhibitor = self.calculate_inhibitors()
        if x_inhibitor > self.maximum_inhibitor_concentration:
            self.evaporator.V = flx.IQ_interpolation(self.inhibitor_objective_function,
                                                     V_min, 1., ytol=1e-4, maxiter=100) 
        else:
            self.check_sugar_concentration()
    
    def update_dilution_water(self, x_titer=None):
        if x_titer is None: x_titer = self.calculate_titer()
        water = self.water_required_to_dilute_to_set_titer(x_titer)
        molar_volume = product.chemicals.Water.V(product.T, 101325) # m3 / mol
        self.dilution_water.imol['Water'] += water / molar_volume / 1000
        
    def water_required_to_dilute_to_set_titer(self, x_titer):
        return (1./x_titer - 1./self.target_titer) * self.product.imass[self.products].sum()
      
        


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module defines the composition_balance function, which performs mass 
energy balance the given oil composition of lipid cane to determine its
composition.

"""
from thermosteam import Stream
from biorefineries import lipidcane as lc
from numpy import array

__all__ = ('set_lipid_fraction',
           'get_lipid_fraction')

def set_lipid_fraction(lipid_fraction, stream=None, data={}):
    """Adjust composition of lipid cane to achieve desired oil fraction (dry weight)."""
    if not stream: stream = lc.lipidcane
    if not data:
        carbs_IDs = ('Glucose', 'Sucrose')
        fiber_IDs = ('Lignin', 'Cellulose', 'Hemicellulose')
        lipid_IDs = ('Lipid',)
        
        carbs   = Stream('Carbs', thermo=stream.thermo)
        fiber   = Stream('Fiber', thermo=stream.thermo)
        lipid   = Stream('Lipid', thermo=stream.thermo)
        carbs.imol[carbs_IDs] = stream.imol[carbs_IDs]
        fiber.imol[fiber_IDs] = stream.imol[fiber_IDs]
        lipid.imol[lipid_IDs] = stream.imol[lipid_IDs]
        
        # Mass property arrays
        carbs_mass = stream.imass[carbs_IDs]
        fiber_mass = stream.imass[fiber_IDs]
        lipid_mass = stream.imass[lipid_IDs]
        
        # Net weight
        carbs_massnet = carbs_mass.sum()
        fiber_massnet = fiber_mass.sum()
        lipid_massnet = lipid_mass.sum()
        
        # Heats of combustion per kg
        LHV_carbs_kg = carbs.LHV/carbs_massnet
        LHV_lipid_kg = lipid.LHV/lipid_massnet
        data['LHV_lipid_over_carbs'] = LHV_lipid_kg / LHV_carbs_kg
        
        
        # Relative composition
        data['r_mass_carbs'] = array(carbs_mass/carbs_massnet)
        data['r_mass_fiber'] = array(fiber_mass/fiber_massnet)
        
    z_mass_lipid = lipid_fraction
    F_mass = stream.F_mass
    z_dry = 0.3
    z_mass_lipid = z_mass_lipid * z_dry 
    z_mass_carbs = 0.149 - z_mass_lipid * data['LHV_lipid_over_carbs']
    z_mass_fiber = z_dry - z_mass_carbs - z_mass_lipid - 0.006 - 0.015
    imass = stream.imass
    imass['Water'] = F_mass*0.7
    imass['Ash'] = F_mass*0.006
    imass['Solids'] = F_mass*0.015
    imass['Lipid'] = z_mass_lipid * F_mass
    imass['Glucose', 'Sucrose'] = data['r_mass_carbs'] * z_mass_carbs * F_mass
    imass['Lignin', 'Cellulose', 'Hemicellulose'] = data['r_mass_fiber'] * z_mass_fiber * F_mass
    if any(stream.mol < 0):
        raise ValueError(f'lipid cane oil composition of {z_mass_lipid*100:.0f}% dry weight is infeasible')

def get_lipid_fraction(stream=None):
    if not stream: stream = lc.lipidcane
    imass = stream.imass
    F_dry_mass = stream.F_mass - imass['Water']
    return imass['Lipid'] / F_dry_mass


# -*- coding: utf-8 -*-
"""
Created on Tue Apr 28 01:25:56 2020

@author: yoelr
"""

from . import nitrogen_generation_package
from . import adiabatic_fixedbed_gas_reactor
from . import surge_tank
from biorefineries.fattyalcohols import units

__all__ = (*nitrogen_generation_package.__all__,
           *adiabatic_fixedbed_gas_reactor.__all__,
           *surge_tank.__all__,
           *units.__all__,
)

from .nitrogen_generation_package import *
from .adiabatic_fixedbed_gas_reactor import *
from .surge_tank import *
from biorefineries.fattyalcohols.units import *

del units